Require Import Coq.Program.Equality.

Section lexicographic_product.

  Variable A : Type. 
  
  Variable B : Type.

  Variable lessA : A -> A -> Prop.

  Variable lessB : B -> B -> Prop.

  Inductive lexprod : (A * B) -> (A * B) -> Prop :=
  | LexprodLeft:  forall a1 b1 a2 b2, lessA a1 a2 -> lexprod (a1, b1) (a2, b2)
  | LexprodRight: forall a b1 b2, lessB b1 b2 -> lexprod (a, b1) (a, b2).

  Lemma lexprod_intro:
    forall a1 b1 a2 b2,
    (lessA a1 a2 \/ a1 = a2) ->
    lessB b1 b2 ->
    lexprod (a1, b1) (a2, b2).
  Proof.
    intros ? ? ? ? ha hb.
    elim ha; intro.
    apply LexprodLeft. auto.
    subst. apply LexprodRight. auto.
  Qed.

  Variable wfA : well_founded lessA.

  Variable wfB : well_founded lessB.

  Lemma lexprod_accessibility:
    forall a,
    Acc lessA a ->
    forall b,
    Acc lessB b ->
    Acc lexprod (a, b).
  Proof.
    induction 1 as [ a _ IHa ]. induction 1 as [ b _ IHb ].
    constructor. intros ? h. dependent destruction h.
    (* LexprodLeft *)
    apply IHa. auto.
    apply wfB.
    (* LexprodRight *)
    apply IHb. auto.
  Qed.

  Lemma lexprod_wf: well_founded lexprod.
  Proof.
    intros [ a b ].
    eauto using lexprod_accessibility.
  Qed.

End lexicographic_product.

